import re

from collections.abc import Iterator
from operator import itemgetter
from typing import Any, Union, cast


# typing.Self added in Python 3.11
try:
    from typing import Self
except ImportError:  # pragma: no cover
    from typing_extensions import Self

# typing.TypeAlias added in Python 3.10
try:
    from typing import TypeAlias
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias

from public import public


COMMASPACE = ', '
SPACE = ' '
IDENTIFIER_RE = r'^[a-zA-Z_][a-zA-Z0-9_]*$'


# The `type` statement was introduced in Python 3.12.
LookupValue: TypeAlias = Union['EnumValue', str]


class EnumMetaclass(type):
    """Meta class for Enums."""

    def __init__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]):
        """Create an Enum class.

        :param cls: The class being defined.
        :param name: The name of the class.
        :param bases: The class's base classes.
        :param namespace: The class attributes.
        """
        super().__init__(name, bases, namespace)
        # values -> EnumValues
        cls._byvalue = {}
        # Figure out if this class has a custom factory for building enum
        # values.  The default is EnumValue, but the class (or one of its
        # bases) can declare a custom one with a special attribute.
        factory = namespace.get('__value_factory__')
        # Figure out the set of enum values on the base classes, to ensure
        # that we don't get any duplicate values.  At the same time, check the
        # base classes for the special attribute.
        for basecls in cls.__mro__:
            if hasattr(basecls, '_byvalue'):
                cls._byvalue.update(basecls._byvalue)  # noqa: SLF001 - Private member accessed: `_byvalue`
            if hasattr(basecls, '__value_factory__'):
                basecls_factory = basecls.__value_factory__
                if factory is not None and basecls_factory != factory:
                    raise TypeError(f'Conflicting enum factory in base class: {basecls_factory}')
                factory = basecls_factory
        # Set the factory default if necessary.
        if factory is None:
            factory = EnumValue
        # For each class attribute, create an enum value and store that back
        # on the class instead of the original value.  Skip Python reserved
        # names.  Also add a mapping from the original value to the enum value
        # instance so we can return the same object on conversion.
        for attr in namespace:
            if not (attr.startswith('__') and attr.endswith('__')):
                value = namespace[attr]
                enumval = factory(cls, value, attr)
                if value in cls._byvalue:
                    other = cls._byvalue[value]
                    # Without this, sort order is undefined and causes
                    # unpredictable results for the test suite.
                    first = attr if attr < other else other
                    second = other if attr < other else attr
                    raise ValueError(
                        f"Conflicting enum value '{value}' for names: '{first}' and '{second}'"
                    )
                # Store as an attribute on the class, and save the attr name.
                setattr(cls, attr, enumval)
                cls._byvalue[value] = attr

    def __dir__(cls) -> list[str]:
        return list(cls._byvalue.values())

    def __repr__(cls) -> str:
        # We want predictable reprs.  Because base Enum items can have any
        # value, the only reliable way to sort the keys for the repr is based
        # on the attribute name, which must be Python identifiers.
        value = COMMASPACE.join(
            f'{value}: {key}' for key, value in sorted(cls._byvalue.items(), key=itemgetter(1))
        )
        return f'<{cls.__name__} {{{value}}}>'

    def __iter__(cls) -> Iterator['EnumValue']:
        for value in cls._byvalue.values():
            yield getattr(cls, value)

    def __getitem__(cls, item: LookupValue) -> 'EnumValue':
        try:
            return getattr(cls, item)  # type: ignore
        except (AttributeError, TypeError):
            if hasattr(item, 'value'):
                attr = cls._byvalue.get(item.value)
                if attr is None:
                    raise KeyError(item) from None
                return cast(EnumValue, getattr(cls, attr))
            missing = object()
            value = cls._byvalue.get(item, missing)
            if value is not missing:
                return cast(EnumValue, getattr(cls, value))
            raise KeyError(item) from None

    def __call__(cls, *args):  # type: ignore
        if len(args) == 1:
            return cls.__getitem__(args[0])
        # Two argument allows for extending enums.
        name, source = args
        return _make(cls, name, source)


@public
class EnumValue:
    """Class representing an enumeration value.

    EnumValue(Color, 'red', 12) prints as 'Color.red' and can be converted
    to the integer 12.
    """

    def __init__(self, enum: EnumMetaclass, value: Any, name: str):
        self._enum = enum
        self._value = value
        self._name = name

    def __repr__(self) -> str:
        return f'<EnumValue: {self._enum.__name__}.{self._name} [value={self._value}]>'

    def __str__(self) -> str:
        return f'{self._enum.__name__}.{self._name}'

    def __reduce__(self):  # type: ignore
        return getattr, (self._enum, self._name)

    @property
    def enum(self) -> EnumMetaclass:
        """The underlying enum."""
        return self._enum

    @property
    def name(self) -> str:
        """The enum's name."""
        return self._name

    @property
    def value(self) -> Any:
        """The enum's value."""
        return self._value

    # Support only comparison by identity and equality.  Ordered comparisons are not supported.
    def __eq__(self, other: object) -> bool:
        return self is other

    def __ne__(self, other: object) -> bool:
        return self is not other

    def __lt__(self, other: Any) -> Any:
        return NotImplemented

    def __gt__(self, other: Any) -> Any:
        return NotImplemented

    def __le__(self, other: Any) -> Any:
        return NotImplemented

    def __ge__(self, other: Any) -> Any:
        return NotImplemented

    __hash__ = object.__hash__


@public
class Enum(metaclass=EnumMetaclass):
    """The public API Enum class."""


class IntEnumValue(int, EnumValue):
    """An EnumValue that is also an integer."""

    def __new__(
        cls,
        enum: EnumMetaclass,  # noqa: ARG003 Unused class method argument
        value: Any,
        attr: str,  # noqa: ARG003 Unused class method argument
    ) -> Self:
        # `attr` is going to be the attribute name as created through the
        # factory call in EnumMetaclass.__init__(), however we need to throw
        # that away when calling int's __new__().
        return super().__new__(cls, value)

    __repr__ = EnumValue.__repr__
    __str__ = EnumValue.__str__

    # The non-deprecated version of this method.
    def __int__(self) -> int:
        return cast(int, self._value)

    __hash__ = int.__hash__

    # For slices and index().
    __index__ = __int__


class IntEnumMetaclass(EnumMetaclass):
    # Define an iteration over the integer values instead of the attribute names.
    def __iter__(cls) -> Iterator[IntEnumValue]:  # noqa: N805 First argument of a method should be `self`
        for key in cls._byvalue:
            yield getattr(cls, cls._byvalue[key])


@public
class IntEnum(metaclass=IntEnumMetaclass):
    """A specialized enumeration with values that are also integers."""

    __value_factory__ = IntEnumValue


def _swap(sequence: Iterator[Any]) -> Iterator[Any]:
    for key, value in sequence:
        yield value, key


def _make(enum_class: EnumMetaclass, name: str, source: Any) -> EnumMetaclass:
    # The common implementation for Enum() and IntEnum().
    namespace = {}
    illegals = []
    have_strings: bool | None = None
    # Auto-splitting of strings.
    if isinstance(source, str):
        source = source.split()
    # Look for dict-like arguments.  Specifically, it must have a callable
    # .items() attribute.  Because of the way enumerate() works, here we have
    # to swap the key/values.
    try:
        source = _swap(source.items())
    except (TypeError, AttributeError):
        source = enumerate(source, start=1)
    for i, item in source:
        if isinstance(item, str):
            if have_strings is None:
                have_strings = True
            elif not have_strings:
                raise ValueError('heterogeneous source')
            namespace[item] = i
            if re.match(IDENTIFIER_RE, item) is None:
                illegals.append(item)
        else:
            if have_strings is None:
                have_strings = False
            elif have_strings:
                raise ValueError('heterogeneous source')
            item_name, item_value = item
            namespace[item_name] = item_value
            if re.match(IDENTIFIER_RE, item_name) is None:
                illegals.append(item_name)
    if len(illegals) > 0:
        raise ValueError(f'non-identifiers: {SPACE.join(illegals)}')
    return EnumMetaclass(str(name), (enum_class,), namespace)



from public import public as _public

from ._enum import Enum, EnumValue, IntEnum


__version__ = '7.0.0'


_public(
    Enum=Enum,
    EnumValue=EnumValue,
    IntEnum=IntEnum,
    __version__=__version__,
)


del _public



import os
import pytest

try:
    # Python < 3.12
    from importlib_resources import files
except ImportError:
    from importlib.resources import files

from contextlib import ExitStack
from sybil import Sybil
from doctest import ELLIPSIS, REPORT_NDIFF, NORMALIZE_WHITESPACE
from sybil.parsers.doctest import DocTestParser
from sybil.parsers.codeblock import PythonCodeBlockParser


DOCTEST_FLAGS = ELLIPSIS | NORMALIZE_WHITESPACE | REPORT_NDIFF


class DoctestNamespace:
    def __init__(self):
        self._resources = ExitStack()

    def setup(self, namespace):
        # The docs include an example of how enums can be pickled and
        # unpickled.  For this, our test module must be importable.  Hack
        # sys.path so that the `fruit.py` module (containing an enum for the
        # pickle tests) can be imported.  There's probably a more robust way to
        # do this, but it works and is simple so fix it only if necessary.
        test_dir = files() / 'tests'
        assert (test_dir / 'fruit.py').is_file()
        self._resources.enter_context(
            pytest.MonkeyPatch.context()).syspath_prepend(test_dir)

    def teardown(self, namespace):
        self._resources.close()


namespace = DoctestNamespace()


pytest_collect_file = Sybil(
    parsers=[
        DocTestParser(optionflags=DOCTEST_FLAGS),
        PythonCodeBlockParser(),
    ],
    pattern='*.rst',
    setup=namespace.setup,
    teardown=namespace.teardown,
).pytest()



import pytest

from flufl.enum import Enum, IntEnum, EnumValue
from flufl.enum._enum import IntEnumValue
from itertools import combinations
from operator import attrgetter, index

SPACE = ' '


class Colors(Enum):
    red = 1
    green = 2
    blue = 3


class OtherColors(Enum):
    red = 1
    blue = 2
    yellow = 3


class MoreColors(Colors):
    pink = 4
    cyan = 5


ALL_COLORS = ['red', 'green', 'blue']


class Rush(Enum):
    geddy = 'bass'
    alex = 'guitar'
    neil = 'drums'


class Animals(IntEnum):
    ant = 1
    bee = 2
    cat = 3


def test_basic_reprs():
    assert str(Colors.red) == 'Colors.red'
    assert str(Colors.green) == 'Colors.green'
    assert str(Colors.blue) == 'Colors.blue'
    assert str(Colors['red']) == 'Colors.red'
    assert repr(Colors.red) == '<EnumValue: Colors.red [value=1]>'


def test_string_value():
    assert repr(Rush.alex) == '<EnumValue: Rush.alex [value=guitar]>'


def test_factory_single_string():
    Color = Enum('Color', SPACE.join(ALL_COLORS))
    for c in ALL_COLORS:
        assert str(Color[c]) == 'Color.' + c


def test_enum_dir():
    # dir() returns the list of enumeration item names.
    assert sorted(dir(Colors)) == sorted(ALL_COLORS)


def test_enumclass_getitem():
    assert Colors[2] is Colors.green
    assert Colors['red'] is Colors.red
    assert Colors[Colors.red] is Colors.red


def test_iteration():
    # Iteration of Enums preserves the order in which the attributes appear.
    A = Enum('A', dict(a=1, b=2, c=3))
    assert [e.name for e in A] == ['a', 'b', 'c']
    B = Enum('B', dict(c=1, b=2, a=3))
    assert [e.name for e in B] == ['c', 'b', 'a']
    # If iteration sorted over values, this would give a TypeError.
    C = Enum('C', dict(a='7', b=7))
    assert [e.name for e in C] == ['a', 'b']


def test_hashing():
    getvalue = attrgetter('value')
    apples = {}
    apples[Colors.red] = 'red delicious'
    apples[Colors.green] = 'granny smith'
    assert [(c.name, apples[c]) for c in sorted(apples, key=getvalue)] == [
        ('red', 'red delicious'),
        ('green', 'granny smith'),
    ]


def test_value_enum_attributes():
    for i, c in enumerate(ALL_COLORS, 1):
        # enum attribute
        assert Colors[c].enum is Colors
        # name attribute
        assert Colors[c].name == c
        # value attribute
        assert Colors[c].value == i


def test_enum_class_name():
    assert Colors.__name__ == 'Colors'


def test_comparisons():
    r, g, b = Colors.red, Colors.green, Colors.blue
    for c in r, g, b:
        assert c is c
        assert c == c
    for first, second in combinations([r, g, b], 2):
        assert first is not second
        assert first != second

    with pytest.raises(TypeError):
        Colors.red < Colors.blue
    with pytest.raises(TypeError):
        Colors.red <= Colors.blue
    with pytest.raises(TypeError):
        Colors.red > Colors.green
    with pytest.raises(TypeError):
        Colors.green >= Colors.blue


def test_comparison_with_int():
    with pytest.raises(TypeError):
        Colors.red < 3
    with pytest.raises(TypeError):
        Colors.red <= 3
    with pytest.raises(TypeError):
        Colors.blue > 2
    with pytest.raises(TypeError):
        Colors.green >= 1

    assert Colors.green != 2
    assert Colors.blue != 3
    assert Colors.green != 3


def test_comparison_with_other_enum():
    assert OtherColors.red is not Colors.red
    assert OtherColors.red != Colors.red
    assert hash(OtherColors.red) != hash(Colors.red)


def test_subclass():
    assert Colors.red is MoreColors.red
    assert Colors.blue is MoreColors.blue


def test_pickle():
    from .fruit import Fruit
    from pickle import dumps, loads

    assert Fruit.tomato is loads(dumps(Fruit.tomato))


def test_functional_api_single_string():
    animals = Enum('Animals', 'ant bee cat dog')
    assert repr(animals) == '<Animals {ant: 1, bee: 2, cat: 3, dog: 4}>'


def test_functional_api_sequence():
    people = Enum('People', ('anne', 'bart', 'cate', 'dave'))
    assert repr(people) == '<People {anne: 1, bart: 2, cate: 3, dave: 4}>'


def test_functional_api_2_tuples():
    def enumiter():
        start = 1
        while True:
            yield start
            start <<= 1

    flags = Enum('Flags', zip(list('abcdefg'), enumiter()))
    assert (
        repr(flags) == '<Flags {a: 1, b: 2, c: 4, d: 8, e: 16, f: 32, g: 64}>'
    )


def test_functional_api_dict():
    # Note: repr is sorted by attribute name
    bassists = dict(geddy='rush', chris='yes', flea='rhcp', jack='cream')
    assert (
        repr(Enum('Bassists', bassists))
        == '<Bassists {chris: yes, flea: rhcp, geddy: rush, jack: cream}>'
    )


def test_invalid_getitem_arguments():
    # Trying to get an invalid value raises an exception.
    with pytest.raises(KeyError) as exc_info:
        Colors['magenta']
    assert exc_info.value.args == ('magenta',)


def test_no_duplicates():
    with pytest.raises(ValueError) as exc_info:

        class Bad(Enum):
            cartman = 1
            stan = 2
            kyle = 3
            kenny = 3   # Oops!
            butters = 4

    assert (
        str(exc_info.value)
        == "Conflicting enum value '3' for names: 'kenny' and 'kyle'"
    )


def test_no_duplicates_in_subclass():
    with pytest.raises(ValueError) as exc_info:

        class BadMoreColors(Colors):
            yellow = 4
            magenta = 2   # Oops!

    assert (
        str(exc_info.value)
        == "Conflicting enum value '2' for names: 'green' and 'magenta'"
    )


def test_no_duplicates_in_dict():
    with pytest.raises(ValueError) as exc_info:
        Enum('Things', dict(a='yes', b='no', c='maybe', d='yes'))
    assert (
        exc_info.value.args[0]
        == "Conflicting enum value 'yes' for names: 'a' and 'd'"
    )


def test_functional_api_not_all_2_tuples():
    # If 2-tuples are used, all items must be 2-tuples.
    with pytest.raises(ValueError):
        Enum(
            'Animals',
            (
                ('ant', 1),
                ('bee', 2),
                'cat',
                ('dog', 4),
            ),
        )
    with pytest.raises(ValueError):
        Enum(
            'Animals',
            (
                ('ant', 1),
                ('bee', 2),
                ('cat',),
                ('dog', 4),
            ),
        )
    with pytest.raises(ValueError):
        Enum(
            'Animals',
            (
                ('ant', 1),
                ('bee', 2),
                ('cat', 3, 'oops'),
                ('dog', 4),
            ),
        )


def test_functional_api_identifiers():
    # Ensure that the functional API enforces identifiers.
    with pytest.raises(ValueError) as exc_info:
        Enum('Foo', ('1', '2', '3'))
    assert exc_info.value.args[0] == 'non-identifiers: 1 2 3'
    with pytest.raises(ValueError) as exc_info:
        Enum('Foo', (('ant', 1), ('bee', 2), ('3', 'cat')))
    assert exc_info.value.args[0] == 'non-identifiers: 3'


def test_functional_api_identifiers_lp1167052():
    # LP: #1167052
    with pytest.raises(ValueError):
        Enum('X', 'a-1')


def test_functional_api_identifiers_numbers():
    # There was a typo in IDENTIFIER_RE where the range 0-0 was used.
    MyEnum = Enum('X', 'a9')
    assert MyEnum.a9.name == 'a9'


def test_explicit_getattr():
    Fruit = Enum('Fruit', 'apple banana tangerine orange')
    assert getattr(Fruit, 'banana') is Fruit.banana
    assert getattr(Fruit, Fruit.banana.name) is Fruit.banana


def test_issue_17576():
    # http://bugs.python.org/issue17576
    #
    # The problem is that despite the documentation, operator.index() is
    # *not* equivalent to calling obj.__index__() when the object in
    # question is an int subclass.
    # Test that while the actual type returned by operator.index() and
    # obj.__index__() are not the same (because the former returns the
    # subclass instance, but the latter returns the .value attribute) they
    # are equal.
    assert index(Animals.bee) == Animals.bee.__index__()


def test_basic_intenum():
    animal_list = [Animals.ant, Animals.bee, Animals.cat]
    assert animal_list == [1, 2, 3]
    assert [int(a) for a in animal_list] == [1, 2, 3]
    assert list(range(10)[Animals.ant : Animals.cat]) == [1, 2]


def test_int_enums_type():
    # IntEnum() enum values are ints.
    Toppings = IntEnum(
        'Toppings',
        dict(olives=1, onions=2, mushrooms=4, cheese=8, garlic=16).items(),
    )
    assert Toppings.garlic == 16
    assert isinstance(Toppings.mushrooms, int)


def test_intenum_comparisons():
    assert Animals.ant < Animals.bee
    assert Animals.cat > Animals.ant
    assert Animals.ant <= 1.0
    assert Animals.bee == 2

    class Toppings(IntEnum):
        anchovies = 1
        black_olives = 2

    assert Animals.bee == Toppings.black_olives


def test_intenum_iteration():
    # Iteration over IntEnums is by value.
    A = IntEnum('A', 'a b c')
    assert list(v.name for v in A) == ['a', 'b', 'c']
    B = IntEnum('B', 'c b a')
    # Iteration over this enum is different than if it were an Enum.
    assert list(v.name for v in B) == ['c', 'b', 'a']


def test_conflicting_factories():
    # An enum extension cannot have a different value factory.
    class First(Enum):
        __value_factory__ = EnumValue
        red = 1
        green = 2
        blue = 3

    with pytest.raises(TypeError):
        class Second(First):
            __value_factory__ = IntEnumValue
            red = 1
            green = 2
            blue = 3


def test_weird_getitem():
    # If we try to get an enum item with an argument that pretends to be an
    # item, but really isn't, we'll get a KeyError.
    class Balls(Enum):
        foot = 1
        base = 2
        golf = 3

    class NotAnEnumItem:
        value = 99

    with pytest.raises(KeyError):
        Balls[NotAnEnumItem()]


def test_make_with_inconsistent_values():
    # Passing a dictionary into the functional API requires that all values be
    # of the same type.
    with pytest.raises(ValueError):
        Enum('Animal', ('ant', 7))
